<?php

namespace Drupal\external_entities\Entity;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableEntityBundleInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\external_entities\Event\ExternalEntitiesEvents;
use Drupal\external_entities\Event\ExternalEntityExtractRawDataEvent;

/**
 * Defines the external entity class.
 *
 * @see external_entities_entity_type_build()
 */
class ExternalEntity extends ContentEntityBase implements ExternalEntityInterface {

  /**
   * Source raw data provided by the storage client.
   *
   * @var array
   */
  protected $rawData;

  /**
   * Returns external entity content entity base definition.
   *
   * @return array
   *   Base content entity definition.
   */
  public static function getBaseDefinition() {
    return [
      'id' => 'external_entity',
      'label' => t('External Entity'),
      'label_plural' => t('External Entities'),
      'label_collection' => t('External Entities'),
      'provider' => 'external_entities',
      'class' => 'Drupal\external_entities\Entity\ExternalEntity',
      'group' => 'content',
      'group_label' => t('Content'),
      'translatable' => TRUE,
      'admin_permission' => 'administer external entity types',
      'permission_granularity' => 'entity_type',
      'persistent_cache' => FALSE,
      'handlers' => [
        'storage' => 'Drupal\external_entities\ExternalEntityStorage',
        'view_builder' => 'Drupal\Core\Entity\EntityViewBuilder',
        'form' => [
          'default' => 'Drupal\external_entities\Form\ExternalEntityForm',
          'edit' => 'Drupal\external_entities\Form\ExternalEntityForm',
          'delete' => 'Drupal\Core\Entity\ContentEntityDeleteForm',
        ],
        'list_builder' => 'Drupal\external_entities\ExternalEntityListBuilder',
        'access' => 'Drupal\external_entities\ExternalEntityAccessControlHandler',
        'route_provider' => [
          'html' => 'Drupal\external_entities\Routing\ExternalEntityHtmlRouteProvider',
        ],
      ],
      'links' => [],
      'entity_keys' => [
        'id' => 'id',
        'uuid' => 'uuid',
        'label' => 'title',
        'langcode' => 'langcode',
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $values,
    $entity_type,
    $bundle = FALSE,
    $translations = [],
  ) {
    $values['langcode'] = $values['langcode'] ?? [
      LanguageInterface::LANGCODE_DEFAULT => LanguageInterface::LANGCODE_NOT_SPECIFIED,
    ];
    parent::__construct($values, $entity_type, $bundle, $translations);
    $this->rawData = [];
  }

  /**
   * {@inheritdoc}
   */
  public function getExternalEntityType() :ExternalEntityTypeInterface {
    return $this
      ->entityTypeManager()
      ->getStorage('external_entity_type')
      ->load($this->getEntityTypeId());
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    return self::defaultBaseFieldDefinitions();
  }

  /**
   * Provides the default base field definitions for external entities.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   An array of base field definitions for the entity type, keyed by field
   *   name.
   */
  public static function defaultBaseFieldDefinitions() {
    $fields = [];

    $fields['id'] = BaseFieldDefinition::create('string')
      ->setLabel(new TranslatableMarkup('ID'))
      ->setReadOnly(TRUE);

    $fields['uuid'] = BaseFieldDefinition::create('string')
      ->setLabel(new TranslatableMarkup('UUID'))
      ->setReadOnly(TRUE);

    $fields['title'] = BaseFieldDefinition::create('string')
      ->setLabel(new TranslatableMarkup('Title'))
      ->setRequired(TRUE)
      ->setDisplayOptions('view', [
        'type' => 'string',
        'label' => 'hidden',
      ])
      ->setTranslatable(TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['langcode'] = BaseFieldDefinition::create('language')
      ->setLabel(new TranslatableMarkup('Language'))
      ->setReadOnly(TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayConfigurable('form', TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public static function bundleFieldDefinitions(EntityTypeInterface $entity_type, $bundle, array $base_field_definitions) {
    $fields = parent::bundleFieldDefinitions($entity_type, $bundle, $base_field_definitions);

    /** @var \Drupal\external_entities\Entity\ExternalEntityTypeInterface $external_entity_type */
    $external_entity_type = \Drupal::entityTypeManager()
      ->getStorage('external_entity_type')
      ->load($entity_type->id());
    if ($external_entity_type && $external_entity_type->isAnnotatable()) {
      // Add the annotation reference field.
      $fields[ExternalEntityInterface::ANNOTATION_FIELD] = BaseFieldDefinition::create('entity_reference')
        ->setLabel(t('Annotation'))
        ->setComputed(TRUE)
        ->setDescription(t('The annotation entity.'))
        ->setSetting('target_type', $external_entity_type->getAnnotationEntityTypeId())
        ->setSetting('handler', 'default')
        ->setSetting('handler_settings', [
          'target_bundles' => [$external_entity_type->getAnnotationBundleId()],
        ]);

      // Have the external entity inherit annotation fields.
      $inherited_fields = $external_entity_type->getInheritedAnnotationFields();
      foreach ($inherited_fields as $field) {
        $field_definition = BaseFieldDefinition::createFromFieldStorageDefinition($field->getFieldStorageDefinition())
          ->setName(ExternalEntityInterface::ANNOTATION_FIELD_PREFIX . $field->getName())
          ->setReadOnly(TRUE)
          ->setComputed(TRUE)
          ->setLabel($field->getLabel())
          ->setDescription($field->getDescription())
          ->setSettings($field->getSettings())
          ->setRequired($field->isRequired())
          ->setTranslatable($field->isTranslatable())
          ->setDefaultValue($field->getDefaultValueLiteral())
          ->setDefaultValueCallback($field->getDefaultValueCallback())
          ->setDisplayConfigurable('form', $field->isDisplayConfigurable('form'))
          ->setDisplayOptions('form', $field->getDisplayOptions('form') ?: [])
          ->setDisplayConfigurable('view', $field->isDisplayConfigurable('view'))
          ->setDisplayOptions('view', $field->getDisplayOptions('view') ?: []);
        $fields[$field_definition->getName()] = $field_definition;
      }
    }

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function toRawData() :array {
    // Not using $this->toArray() here because we don't want computed values.
    $entity_values = [];
    foreach ($this->getFields(FALSE) as $name => $property) {
      $entity_values[$name] = $property->getValue();
    }

    $storage = $this->entityTypeManager()->getStorage($this->getEntityTypeId());
    $raw_data = $storage->createRawDataFromEntityValues(
      $entity_values,
      $this->getOriginalRawData()
    );

    // Allow other modules to perform custom extraction logic.
    $event = new ExternalEntityExtractRawDataEvent($this, $raw_data);
    \Drupal::service('event_dispatcher')->dispatch(
      $event,
      ExternalEntitiesEvents::EXTRACT_RAW_DATA
    );

    return $event->getRawData();
  }

  /**
   * {@inheritdoc}
   */
  public function setOriginalRawData(array $raw_data) :self {
    $this->rawData = $raw_data;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getOriginalRawData() :array {
    return $this->rawData;
  }

  /**
   * {@inheritdoc}
   */
  public function createAnnotation(array $values = []) :ContentEntityInterface {
    $external_entity_type = $this->getExternalEntityType();
    $annotation_entity_type = $this
      ->entityTypeManager()
      ->getDefinition($external_entity_type->getAnnotationEntityTypeId());

    $bundle_key = $annotation_entity_type->getKey('bundle');
    if ($bundle_key) {
      $values[$bundle_key] = $external_entity_type->getAnnotationBundleId();
    }

    $langcode_key = $annotation_entity_type->getKey('langcode');
    if ($langcode_key) {
      $values[$langcode_key] = $this->getUntranslated()->language()->getId();
    }

    return $this
      ->entityTypeManager()
      ->getStorage($external_entity_type->getAnnotationEntityTypeId())
      ->create($values + [
        // @todo What do we do if the external entity doesn't have an ID?
        $external_entity_type->getAnnotationFieldName() => $this->id(),
      ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotation() :ContentEntityInterface|null {
    if (!$this->id()) {
      return NULL;
    }

    $annotation = current(
      $this
        ->entityTypeManager()
        ->getStorage($this->getEntityTypeId())
        ->getAnnotations([$this->id()])
    );

    return $annotation ?: NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function mapAnnotationFields(?ContentEntityInterface $annotation = NULL) :ExternalEntityInterface {
    $external_entity_type = $this->getExternalEntityType();
    if (!$external_entity_type->isAnnotatable()) {
      return $this;
    }

    if (!$annotation) {
      $annotation = $this->getAnnotation();
    }
    if (!$annotation) {
      return $this;
    }

    $this->set(ExternalEntityInterface::ANNOTATION_FIELD, $annotation);
    $inherited_fields = $external_entity_type->getInheritedAnnotationFields();
    foreach (array_keys($this->getTranslationLanguages()) as $langcode) {
      if ($annotation instanceof TranslatableInterface && $annotation->isTranslatable()) {
        if (!$annotation->hasTranslation($langcode)) {
          continue;
        }

        $annotation_translation = $annotation->getTranslation($langcode);
      }
      else {
        if ($this->getUntranslated()->language()->getId() !== $langcode) {
          continue;
        }

        $annotation_translation = $annotation;
      }

      foreach ($inherited_fields as $field_name => $inherited_field) {
        $this
          ->getTranslation($langcode)
          ->set(
            ExternalEntityInterface::ANNOTATION_FIELD_PREFIX . $field_name,
            $annotation_translation->get($field_name)->getValue(),
          );
      }
    }

    return $this;
  }

  /**
   * Gets the fields that can be inherited by the external entity.
   *
   * @param \Drupal\external_entities\Entity\ExternalEntityTypeInterface $type
   *   The type of the external entity.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   An array of field definitions, keyed by field name.
   *
   * @see \Drupal\Core\Entity\EntityManagerInterface::getFieldDefinitions()
   */
  public static function getInheritedAnnotationFields(ExternalEntityTypeInterface $type) {
    $inherited_fields = [];

    $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions($type->getAnnotationEntityTypeId(), $type->getAnnotationBundleId());
    foreach ($field_definitions as $field_name => $field_definition) {
      if ($field_name !== $type->getAnnotationFieldName()) {
        $inherited_fields[$field_name] = $field_definition;
      }
    }

    return $inherited_fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() {
    return Cache::mergeTags(
      Cache::mergeTags(parent::getCacheTags(), $this->getCacheTagsToInvalidate()),
      $this->getExternalEntityType()->getCacheTags()
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTagsToInvalidate() {
    return Cache::mergeTags(
      parent::getCacheTagsToInvalidate(),
      $this->getExternalEntityType()->getCacheTagsToInvalidate()
    );
  }

  /**
   * {@inheritDoc}
   */
  public function getCacheMaxAge() {
    return $this->getExternalEntityType()->getPersistentCacheMaxAge();
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE) :void {
    parent::postSave($storage, $update);

    $this->saveAnnotationAfterExternalEntitySave($storage);
  }

  /**
   * Saves the annotation after the external entity is saved.
   *
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage object.
   *
   * @see ExternalEntity::postSave()
   */
  protected function saveAnnotationAfterExternalEntitySave(EntityStorageInterface $storage) :void {
    if (!$this->hasField(ExternalEntityInterface::ANNOTATION_FIELD)) {
      return;
    }

    // If this external entity was automatically saved because of a change in
    // the annotation, there's no need to save the annotation again.
    if (!empty($this->{ExternalEntityInterface::EXTERNAL_ENTITY_AUTO_SAVE_INDUCED_BY_ANNOTATION_CHANGE_PROPERTY})) {
      return;
    }

    /** @var \Drupal\Core\Entity\ContentEntityInterface|null $annotation */
    $annotation = $this
      ->get(ExternalEntityInterface::ANNOTATION_FIELD)
      ->entity;
    if (!$annotation) {
      $annotation = $this->createAnnotation();
      $this->set(ExternalEntityInterface::ANNOTATION_FIELD, $annotation);
    }
    if (!$annotation) {
      return;
    }

    // Check if a new revision needs to be created or not.
    if ($annotation instanceof RevisionableInterface) {
      $bundle_entity_type = $annotation
        ->getEntityType()
        ->getBundleEntityType();
      if ($bundle_entity_type) {
        $bundle_entity = $this
          ->entityTypeManager()
          ->getStorage($bundle_entity_type)
          ->load($annotation->bundle());
        if ($bundle_entity instanceof RevisionableEntityBundleInterface && $bundle_entity->shouldCreateNewRevision()) {
          $annotation->setNewRevision();
        }
      }
    }

    // If the annotation is translatable we make sure all necessary translations
    // are present on the annotation.
    $new_translations = [];
    if ($annotation instanceof TranslatableInterface && $annotation->isTranslatable()) {
      foreach (array_keys($this->getTranslationLanguages()) as $langcode) {
        if (!$annotation->hasTranslation($langcode)) {
          $annotation->addTranslation($langcode);
          $new_translations[] = $langcode;
        }
      }
    }

    $save_annotation = FALSE;
    foreach (array_keys($this->getTranslationLanguages()) as $langcode) {
      if ($annotation instanceof TranslatableInterface && $annotation->isTranslatable()) {
        $annotation_translation = $annotation->getTranslation($langcode);
      }
      else {
        if ($this->getUntranslated()->language()->getId() !== $langcode) {
          continue;
        }
        $annotation_translation = $annotation;
      }

      foreach ($this->getExternalEntityType()->getInheritedAnnotationFields() as $field_definition) {
        $inherited_field_name = ExternalEntityInterface::ANNOTATION_FIELD_PREFIX . $field_definition->getName();
        if (!$this->hasField($inherited_field_name) || !$annotation->hasField($field_definition->getName())) {
          continue;
        }

        if (!$annotation_translation->get($field_definition->getName())->equals($this->getTranslation($langcode)->get($inherited_field_name))) {
          if (!$annotation->isNew() || !$this->getTranslation($langcode)->get($inherited_field_name)->isEmpty()) {
            $annotation_translation->set($field_definition->getName(), $this->getTranslation($langcode)->get($inherited_field_name)->getValue());
          }
          else {
            $this->getTranslation($langcode)->set($inherited_field_name, $annotation_translation->get($field_definition->getName())->getValue());
          }
          $save_annotation = TRUE;
        }
        elseif ($annotation->isNew() || in_array($langcode, $new_translations, TRUE)) {
          if (!$this->getTranslation($langcode)->get($inherited_field_name)->isEmpty()) {
            $annotation_translation->set($field_definition->getName(), $this->getTranslation($langcode)->get($inherited_field_name)->getValue());
            $save_annotation = TRUE;
          }
        }
      }
    }

    if ($save_annotation) {
      $annotation->{ExternalEntityInterface::ANNOTATION_AUTO_SAVE_INDUCED_BY_EXTERNAL_ENTITY_CHANGE_PROPERTY} = TRUE;
      $annotation->save();
      $storage->resetCache([$this->id()]);
    }
  }

}
